@@ -2,24 +2,14 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 
-from __future__ import absolute_import, unicode_literals, print_function
-
 import os
 import re
 import sys
 from argparse import Namespace
 
-from mozbuild.base import (
-    MachCommandBase,
-    MachCommandConditions as conditions,
-    MozbuildObject,
-)
-
-from mach.decorators import (
-    CommandProvider,
-    Command,
-)
-
+from mach.decorators import Command
+from mozbuild.base import MachCommandConditions as conditions
+from mozbuild.base import MozbuildObject
 
 parser = None
 
@@ -156,7 +146,7 @@ class ReftestRunner(MozbuildObject):
         if not args.xrePath:
             args.xrePath = os.environ.get("MOZ_HOST_BIN")
         if not args.app:
-            args.app = "org.mozilla.geckoview.test"
+            args.app = "org.mozilla.geckoview.test_runner"
         if not args.utilityPath:
             args.utilityPath = args.xrePath
         args.ignoreWindowSize = True
@@ -170,6 +160,10 @@ class ReftestRunner(MozbuildObject):
             args.e10s = False
             print("using e10s=False for non-geckoview app")
 
+        # Disable fission until geckoview supports fission by default.
+        # Need fission on Android? Use '--setpref fission.autostart=true'
+        args.disableFission = True
+
         # A symlink and some path manipulations are required so that test
         # manifests can be found both locally and remotely (via a url)
         # using the same relative path.
@@ -228,77 +222,76 @@ def get_parser():
     return parser
 
 
-@CommandProvider
-class MachCommands(MachCommandBase):
-    @Command(
-        "reftest",
-        category="testing",
-        description="Run reftests (layout and graphics correctness).",
-        parser=get_parser,
-    )
-    def run_reftest(self, command_context, **kwargs):
-        kwargs["suite"] = "reftest"
-        return self._run_reftest(**kwargs)
+@Command(
+    "reftest",
+    category="testing",
+    description="Run reftests (layout and graphics correctness).",
+    parser=get_parser,
+)
+def run_reftest(command_context, **kwargs):
+    kwargs["suite"] = "reftest"
+    return _run_reftest(command_context, **kwargs)
 
-    @Command(
-        "jstestbrowser",
-        category="testing",
-        description="Run js/src/tests in the browser.",
-        parser=get_parser,
-    )
-    def run_jstestbrowser(self, command_context, **kwargs):
-        if "--enable-js-shell" not in self.mozconfig["configure_args"]:
-            raise Exception(
-                "jstestbrowser requires --enable-js-shell be specified in mozconfig."
-            )
-        self._mach_context.commands.dispatch(
-            "build", self._mach_context, what=["stage-jstests"]
+
+@Command(
+    "jstestbrowser",
+    category="testing",
+    description="Run js/src/tests in the browser.",
+    parser=get_parser,
+)
+def run_jstestbrowser(command_context, **kwargs):
+    if command_context.substs.get("JS_DISABLE_SHELL"):
+        raise Exception(
+            "jstestbrowser requires --enable-js-shell be specified in mozconfig."
         )
-        kwargs["suite"] = "jstestbrowser"
-        return self._run_reftest(**kwargs)
-
-    @Command(
-        "crashtest",
-        category="testing",
-        description="Run crashtests (Check if crashes on a page).",
-        parser=get_parser,
+    command_context._mach_context.commands.dispatch(
+        "build", command_context._mach_context, what=["stage-jstests"]
     )
-    def run_crashtest(self, command_context, **kwargs):
-        kwargs["suite"] = "crashtest"
-        return self._run_reftest(**kwargs)
+    kwargs["suite"] = "jstestbrowser"
+    return _run_reftest(command_context, **kwargs)
 
-    def _run_reftest(self, **kwargs):
-        kwargs["topsrcdir"] = self.topsrcdir
-        process_test_objects(kwargs)
-        reftest = self._spawn(ReftestRunner)
-        # Unstructured logging must be enabled prior to calling
-        # adb which uses an unstructured logger in its constructor.
-        reftest.log_manager.enable_unstructured()
-        if conditions.is_android(self):
-            from mozrunner.devices.android_device import (
-                verify_android_device,
-                InstallIntent,
-            )
 
-            install = (
-                InstallIntent.NO if kwargs.get("no_install") else InstallIntent.YES
-            )
-            verbose = False
-            if (
-                kwargs.get("log_mach_verbose")
-                or kwargs.get("log_tbpl_level") == "debug"
-                or kwargs.get("log_mach_level") == "debug"
-                or kwargs.get("log_raw_level") == "debug"
-            ):
-                verbose = True
-            verify_android_device(
-                self,
-                install=install,
-                xre=True,
-                network=True,
-                app=kwargs["app"],
-                device_serial=kwargs["deviceSerial"],
-                verbose=verbose,
-            )
-            return reftest.run_android_test(**kwargs)
-        return reftest.run_desktop_test(**kwargs)
+@Command(
+    "crashtest",
+    category="testing",
+    description="Run crashtests (Check if crashes on a page).",
+    parser=get_parser,
+)
+def run_crashtest(command_context, **kwargs):
+    kwargs["suite"] = "crashtest"
+    return _run_reftest(command_context, **kwargs)
+
+
+def _run_reftest(command_context, **kwargs):
+    kwargs["topsrcdir"] = command_context.topsrcdir
+    process_test_objects(kwargs)
+    reftest = command_context._spawn(ReftestRunner)
+    # Unstructured logging must be enabled prior to calling
+    # adb which uses an unstructured logger in its constructor.
+    reftest.log_manager.enable_unstructured()
+    if conditions.is_android(command_context):
+        from mozrunner.devices.android_device import (
+            InstallIntent,
+            verify_android_device,
+        )
+
+        install = InstallIntent.NO if kwargs.get("no_install") else InstallIntent.YES
+        verbose = False
+        if (
+            kwargs.get("log_mach_verbose")
+            or kwargs.get("log_tbpl_level") == "debug"
+            or kwargs.get("log_mach_level") == "debug"
+            or kwargs.get("log_raw_level") == "debug"
+        ):
+            verbose = True
+        verify_android_device(
+            command_context,
+            install=install,
+            xre=True,
+            network=True,
+            app=kwargs["app"],
+            device_serial=kwargs["deviceSerial"],
+            verbose=verbose,
+        )
+        return reftest.run_android_test(**kwargs)
+    return reftest.run_desktop_test(**kwargs)
